﻿using System.Collections;
using System.Collections.Generic;
using System.Linq;
using UnityEngine;
using UnityEngine.UI;

public class SwimmerManager : MonoBehaviour
{
    public GameObject swimmerPrefab;
    public List<GameObject> swimmers;
    public int initialSwimmerCount;
    public int maxOffscreenSwimmerCount;

    public Text savedScoreText;
    public Text lostScoreText;

    private int savedScore;
    private int lostScore;

    [Range(1,30)]
    public float struggleTimerMin;
    [Range(1,30)]
    public float struggleTimerMax;
    [Range(1,10)]
    public float drownTime;

    public float spawnDelaySecs;

    public static SwimmerManager Instance;

    private GameObjectPool swimmerPool;
    private float swimmerSize;

    void Awake() {
      if (SwimmerManager.Instance) DestroyImmediate(this);
      SwimmerManager.Instance = this;
      swimmerPool = new GameObjectPool(swimmerPrefab);
    }

    void Start() {
      foreach (var obj in GameObject.FindGameObjectsWithTag("Swimmer")) {
        // Destroy(obj);
      }
      swimmers = new List<GameObject>();

      StartCoroutine(Spawner());
      SpawnSwimmerAt(Vector3.zero);

      Reset();
    }

    void Reset() {
      savedScore = 0;
      lostScore = 0;
    }

    void Update() {
      savedScoreText.text = "" + savedScore;
      lostScoreText.text = "" + lostScore;

      // TODO: move this to a game manager
      if (Input.GetKeyDown(KeyCode.Escape)) {
        Application.Quit();
      }
    }

    public void SwimmerLost(GameObject swimmer) {
      lostScore++;
      swimmers.Remove(swimmer);
      swimmerPool.Destroy(swimmer);

    }

    public void SwimmerSaved(GameObject swimmer) {
      savedScore++;
      swimmers.Remove(swimmer);
      swimmerPool.Destroy(swimmer);
    }

    IEnumerator Spawner() {
      // Initial spawn on screen
      var viewport = new ViewportWorldCoords();
      // Debug.Log(string.Format("Bounds: {0}, {1}, {2}, {3}", viewport.bounds[0], viewport.bounds[1], viewport.bounds[2], viewport.bounds[3]));
      for (var i = 0; i < initialSwimmerCount; i++) {
        var corner = viewport.bounds[i % 4];
        var spawnPoint = new Vector3(Random.Range(0f, corner.x), Random.Range(0f, corner.y), 0f);
        // Debug.Log(string.Format("Spawning at: ({0}, {1})", spawnPoint.x, spawnPoint.y));
        SpawnSwimmerAt(spawnPoint);
      }

      // Continual spawning
      var delay = new WaitForSeconds(spawnDelaySecs);
      while(true) {
        yield return delay;
        var invisibleCount = swimmers.Aggregate(0, (count, swimmer) => {
          if (swimmer.GetComponentInChildren<SpriteRenderer>().isVisible) {
            count++;
          }
          return count;
        });
        if (invisibleCount >= maxOffscreenSwimmerCount) continue;
        SpawnSwimmerOffscreen();
      }
    }

    // Spawns swimmers in a circle that circumscribes the viewport rectangle.
    void SpawnSwimmerOffscreen() {
      var viewport = new ViewportWorldCoords();
      float spawnDistance = Mathf.Max(viewport.topRight.x, viewport.topRight.y) + swimmerSize;
      Vector3 spawnPoint = Random.insideUnitCircle * spawnDistance;
      SpawnSwimmerAt(spawnPoint);
    }

    void SpawnSwimmerAt(Vector3 spawnPoint) {
      var newSwimmer = GetSwimmer(spawnPoint);

      var swimRenderer = newSwimmer.GetComponentInChildren<SpriteRenderer>();
      swimmerSize = swimRenderer.bounds.size.magnitude;

      swimmers.Add(newSwimmer);
    }

    GameObject GetSwimmer() {
      var screenWidth = Camera.main.orthographicSize * Screen.width / Screen.height;
      var screenHeight = Camera.main.orthographicSize * Screen.height / Screen.width;
      var spawnPoint = new Vector3(Random.Range(-screenWidth/2f, screenWidth/2f), Random.Range(-screenHeight/2f, screenHeight/2f));
      return GetSwimmer(spawnPoint);
    }

    GameObject GetSwimmer(Vector3 spawnPoint) {
      var newSwimmer = swimmerPool.Instantiate(swimmerPrefab);
      newSwimmer.transform.position = spawnPoint;
      newSwimmer.transform.parent = transform;

      var swimController = newSwimmer.GetComponent<SwimmerController>();
      swimController.struggleTimer = Random.Range(struggleTimerMin, struggleTimerMax);
      swimController.drownTimer = drownTime;

      return newSwimmer;
    }

    private class ViewportWorldCoords {
      public Vector3 topLeft      { get { return Camera.main.ViewportToWorldPoint(viewportTopLeft);     } }
      public Vector3 topRight     { get { return Camera.main.ViewportToWorldPoint(viewportTopRight);    } }
      public Vector3 bottomRight  { get { return Camera.main.ViewportToWorldPoint(viewportBottomRight); } }
      public Vector3 bottomLeft   { get { return Camera.main.ViewportToWorldPoint(viewportBottomLeft);  } }
      public List<Vector3> bounds { get { return new List<Vector3>() { topLeft, topRight, bottomRight, bottomLeft }; } }

      private Vector3 viewportTopLeft     { get { return new Vector3(0, 1, -Camera.main.transform.position.z); } }
      private Vector3 viewportTopRight    { get { return new Vector3(1, 1, -Camera.main.transform.position.z); } }
      private Vector3 viewportBottomRight { get { return new Vector3(1, 0, -Camera.main.transform.position.z); } }
      private Vector3 viewportBottomLeft  { get { return new Vector3(0, 0, -Camera.main.transform.position.z); } }
    }

}

